import numpy as np
from topsy import cell_layout

def test_randomization():
    offsets = np.array([0, 10, 30])
    lengths = np.array([10, 20, 20])
    centres = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [2.0,0.0,0.0]])
    cl = cell_layout.CellLayout(centres, offsets, lengths)

    order = cl.randomize_within_cells()
    assert (order[:10]<10).all()
    assert ((order[10:30]<30) & (order[10:30]>=10)).all()
    assert ((order[30:50]<50) & (order[30:50]>=30)).all()

    assert (order != np.arange(50)).any()

def test_from_positions():
    Npart = 10000
    Nside = 10
    Ncells_to_test = 100
    min_pos = -1.0
    max_pos = 1.0

    np.random.seed(1337)

    pos = np.random.uniform(min_pos, max_pos, (Npart, 3))

    cl, order = cell_layout.CellLayout.from_positions(pos, min_pos, max_pos, Nside)

    pos = pos[order]

    for test_cell in np.random.randint(0, Nside**3, Ncells_to_test):

        # get the cell slice
        cell_slice = cl.cell_slice(test_cell)

        # get the positions of the particles in the cell
        test_pos = pos[cell_slice]

        # get the cell centre
        cell_centre = cl._centres[test_cell]

        cell_size = (max_pos - min_pos) / Nside

        # check that all the particles are within the cell
        assert ((test_pos > (cell_centre - 0.5*cell_size)) & (test_pos < (cell_centre + 0.5*cell_size))).all()






import logging
import numpy as np
import wgpu

from . import loader, split_buffers

logger = logging.getLogger(__name__)


class ParticleBuffers:
    def __init__(self, loader: loader.AbstractDataLoader, device: wgpu.GPUDevice, max_draw_calls_per_buffer: int):
        self.buffers = {}
        self._split_buffers = split_buffers.SplitBuffers(len(loader))
        self._device = device
        self._loader = loader

        self.quantity_name = None
        self._named_quantity_buffers = None
        self._quantity_buffer_is_for_name = None
        self._current_vertex_buffers = []

        self._create_indirect_draw_buffers(max_draw_calls_per_buffer)

        self._last_bufnum = -1

    def _create_indirect_draw_buffers(self, max_draw_calls_per_buffer: int):
        self._indirect_buffers = [] # for indirect draw calls, one needed per physical buffer
        self._indirect_count_buffers = []
        self._indirect_buffers_npy = []
        self._indirect_count_buffers_npy = []
        self._max_draw_calls_per_buffer = max_draw_calls_per_buffer

        for i in range(self._split_buffers.num_buffers):
            self._indirect_buffers.append(
                self._device.create_buffer(size=max_draw_calls_per_buffer * 4 * np.dtype(np.uint32).itemsize,
                                           usage = wgpu.BufferUsage.INDIRECT | wgpu.BufferUsage.COPY_DST)
            )
            #self._indirect_count_buffers.append(
            #    self._device.create_buffer(size=np.dtype(np.uint32).itemsize,
            #                               usage=wgpu.BufferUsage.INDIRECT | wgpu.BufferUsage.COPY_DST)
            #)
            self._indirect_buffers_npy.append(np.zeros((max_draw_calls_per_buffer, 4), dtype=np.uint32))
            #self._indirect_count_buffers_npy.append(np.zeros((1,), dtype=np.uint32))
            self._indirect_buffers_npy[-1][:, 0] = 6 # vertex count


    def specify_vertex_buffer_assignment(self, buffer_names):
        buffers = []
        for name in buffer_names:
            match name:
                case "pos_smooth":
                    buffers.append(self.get_pos_smooth_buffers())
                case "mass":
                    buffers.append(self.get_mass_buffers())
                case "quantity":
                    buffers.append(self.get_quantity_buffers())
                case "rgb_masses":
                    buffers.append(self.get_rgb_masses_buffers())
                case _:
                    raise ValueError(f"Unknown buffer name: {name}")
        self._current_vertex_buffers = buffers
        self._last_bufnum = -1

    def set_vertex_buffers(self, bufnum: int, render_pass: wgpu.GPURenderPassEncoder):
        if bufnum == self._last_bufnum:
            return
        for i, buffers in enumerate(self._current_vertex_buffers):
            render_pass.set_vertex_buffer(i, buffers[bufnum])
        self._last_bufnum = bufnum

    def issue_draw_indirect(self, sph_render_pass: wgpu.GPURenderPassEncoder):

        for bufnum in range(self._split_buffers.num_buffers):
            self.set_vertex_buffers(bufnum, sph_render_pass)
            sph_render_pass._multi_draw_indirect(self._indirect_buffers[bufnum], 0, self._max_draw_calls_per_buffer)

    def update_particle_ranges(self, particle_mins: list[int], particle_lens: list[int]):
        per_buf_start_lens = self._split_buffers.global_to_split_monotonic(particle_mins, particle_lens)
        for bufnum, (particle_min, particle_len) in enumerate(per_buf_start_lens):
            self._indirect_buffers_npy[bufnum][len(particle_min):,1] = 0
            self._indirect_buffers_npy[bufnum][:len(particle_min), 1] = particle_len # instance count
            self._indirect_buffers_npy[bufnum][:len(particle_min), 3] = particle_min # first instance
            self._device.queue.write_buffer(self._indirect_buffers[bufnum], 0, self._indirect_buffers_npy[bufnum])

    def get_pos_smooth_buffers(self):
        if not hasattr(self, "_pos_smooth_buffers"):
            logger.info("Creating position+smoothing buffer")
            data = self._loader.get_pos_smooth().astype(np.float32)
            self._pos_smooth_buffers = self._split_buffers.create_buffers(self._device, 4 * 4,
                                                                          wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.STORAGE | wgpu.BufferUsage.COPY_DST)
            self._split_buffers.write_buffers(self._device, self._pos_smooth_buffers, data)
        return self._pos_smooth_buffers

    def get_mass_buffers(self):
        if not hasattr(self, "_mass_buffers"):
            logger.info("Creating mass buffer")
            data = self._loader.get_mass().astype(np.float32)
            self._mass_buffers = self._split_buffers.create_buffers(self._device, 4,
                                                                     wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.STORAGE | wgpu.BufferUsage.COPY_DST)
            self._split_buffers.write_buffers(self._device, self._mass_buffers, data)
        return self._mass_buffers

    def get_quantity_buffers(self):
        if self.quantity_name is None:
            return self.get_mass_buffers()
        elif self._quantity_buffer_is_for_name != self.quantity_name:
            self._create_quantity_buffers_if_needed()
            logger.info(f"Transferring {self.quantity_name} into buffer")
            data = self._loader.get_named_quantity(self.quantity_name).view(np.float32)
            self._split_buffers.write_buffers(self._device, self._named_quantity_buffers, data)
            self._quantity_buffer_is_for_name = self.quantity_name
        return self._named_quantity_buffers

    def get_rgb_masses_buffers(self):
        if not hasattr(self, "_rgb_masses_buffers"):
            logger.info("Creating RGB masses buffer")
            data = self._loader.get_rgb_masses().view(np.float32)
            self._rgb_masses_buffers = self._split_buffers.create_buffers(self._device, 4 * 3,
                                                wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.STORAGE | wgpu.BufferUsage.COPY_DST)
            self._split_buffers.write_buffers(self._device, self._rgb_masses_buffers, data)
        return self._rgb_masses_buffers

    def _create_quantity_buffers_if_needed(self):
        if self._named_quantity_buffers is not None:
            return
        logger.info("Creating quantity buffer")
        self._named_quantity_buffers = self._split_buffers.create_buffers(self._device, 4,
                                             wgpu.BufferUsage.VERTEX | wgpu.BufferUsage.STORAGE | wgpu.BufferUsage.COPY_DST)


from __future__ import annotations

from typing import Any, Dict, Union, Callable

from PySide6 import QtWidgets, QtCore
from superqt import QLabeledDoubleRangeSlider, QLabeledDoubleSlider
from rendercanvas import BaseRenderCanvas

from .lineedit import MyLineEdit

import math
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

class MapControlsBase(QtWidgets.QDialog):
    def __init__(self, parent: BaseRenderCanvas):
        self._parent = parent
        super().__init__()

        self.setWindowTitle("Color controls")
        self.setWindowFlags(QtCore.Qt.WindowType.Popup | QtCore.Qt.WindowType.FramelessWindowHint)
        self.resize(400, 0)

    def open(self):
        action_rect = self._parent._toolbar.actionGeometry(self._parent._open_cmap) # EEK!
        popoverPosition = self._parent._toolbar.mapToGlobal(action_rect.topLeft())
        super().show()
        self.move(popoverPosition - QtCore.QPoint(self.width()//2, self.height()))


class QLabeledDoubleRangeSliderWithAutoscale(QLabeledDoubleRangeSlider):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._scale_exponent = 0
        super().__init__(*args, **kwargs)

    def _scale_float(self, value: float) -> float:
        return value / 10**self._scale_exponent

    def _unscale_float(self, value: float) -> float:
        return value * 10**self._scale_exponent

    def _repr_value_to_scale_exponent(self, value: float) -> int:
        if value == 0.0:
            return 0
        try:
            exponent = math.floor(math.log10(abs(value)))
        except ValueError:
            return 0
        if exponent < -2 or exponent > 2:
            return exponent
        else:
            return 0

    def setRange(self, vmin: float, vmax: float) -> None:
        if vmin == 0.0 and vmax == 0.0:
            repr_val = 1.0
        elif vmin==0.0:
            repr_val = vmax
        elif vmax==0.0:
            repr_val = vmin
        else:
            repr_val = max(abs(vmin), abs(vmax))

        self._scale_exponent = self._repr_value_to_scale_exponent(repr_val)
        scaled_min = self._scale_float(vmin)
        scaled_max = self._scale_float(vmax)

        super().setRange(scaled_min, scaled_max)

    def setValue(self, value: tuple[float, float]) -> None:
        scaled_value = (self._scale_float(value[0]), self._scale_float(value[1]))
        super().setValue(scaled_value)

    def value(self) -> tuple[float, float]:
        scaled_value = super().value()
        return (self._unscale_float(scaled_value[0]), self._unscale_float(scaled_value[1]))



from ..generic_colormap import (
    GenericController, ColorMapController, RGBMapController, ControlSpec, LayoutSpec
)

class ColorMapControlsBase(QtWidgets.QDialog):
    controller_class : GenericController = None

    def __init__(self, canvas: BaseRenderCanvas):
        super().__init__(canvas)
        self.setWindowTitle("Color controls")
        self.setWindowFlags(QtCore.Qt.WindowType.Popup
                            | QtCore.Qt.WindowType.FramelessWindowHint)
        self.resize(400, 0)

        self.controller = self.controller_class(canvas._visualizer, self._refresh_ui)
        # build UI
        self._widgets: Dict[str, QtWidgets.QWidget] = {}
        root_spec = self.controller.get_layout()
        self.setLayout(self._build_layout(root_spec))

    def open(self):
        # position next to toolbar
        action_rect = self.parent()._toolbar.actionGeometry(
            self.parent()._open_cmap
        )
        pos = self.parent()._toolbar.mapToGlobal(action_rect.topLeft())
        self._refresh_ui()
        super().show()
        self.move(pos - QtCore.QPoint(self.width()//2, self.height()))

    def _build_layout(self, spec: LayoutSpec) -> QtWidgets.QLayout:
        if spec.type == "vbox":
            layout = QtWidgets.QVBoxLayout()
        else:
            layout = QtWidgets.QHBoxLayout()

        for child in spec.children:
            if isinstance(child, ControlSpec):
                label_layout = self._make_widget(child)
                self._widgets[child.name] = label_layout

                if child.label is not None and child.type != "button" and child.type != "checkbox":
                    w_inner = label_layout
                    label_layout = QtWidgets.QHBoxLayout()
                    label_layout.addWidget(QtWidgets.QLabel(child.label))
                    label_layout.addWidget(w_inner)
                    label_layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
                    layout.addLayout(label_layout)
                else:
                    layout.addWidget(label_layout)
            else:
                layout.addLayout(self._build_layout(child))
        return layout

    def _make_widget(self, spec: ControlSpec) -> QtWidgets.QWidget:
        if spec.type == "combo" or spec.type == "combo-edit":
            w = QtWidgets.QComboBox()
            w.setEditable(spec.name == "quantity")

            w.addItems(spec.options or [])
            w.setCurrentText(spec.value)
            edited_callback = lambda: self._on_changed(spec.callback, w.currentText())
            w.currentIndexChanged.connect(edited_callback)
            if spec.type == "combo-edit":
                w.setLineEdit(MyLineEdit())
                w.lineEdit().editingFinished.connect(edited_callback)

        elif spec.type == "checkbox":
            w = QtWidgets.QCheckBox(spec.label or "")
            w.setChecked(bool(spec.value))
            w.stateChanged.connect(
                lambda st, cb=spec.callback: self._on_changed(cb, bool(st))
            )
        elif spec.type == "range_slider":
            w = QLabeledDoubleRangeSliderWithAutoscale()
            w.setRange(*(spec.range or (0.0, 1.0)))
            w.setValue(tuple(spec.value))
            w.valueChanged.connect(
                lambda _, cb=spec.callback, widget=w:
                    self._on_changed(cb, widget.value())
            )
        elif spec.type == "slider":
            w = QLabeledDoubleSlider()
            w.setRange(*(spec.range or (0.0, 1.0)))
            w.setValue(spec.value)
            w.valueChanged.connect(
                lambda _, cb=spec.callback, widget=w:
                    self._on_changed(cb, widget.value())
            )
        elif spec.type == "button":
            w = QtWidgets.QPushButton(spec.label or "")
            w.setStyleSheet("color: black;")  # unclear why this is necessary
            w.pressed.connect(lambda cb=spec.callback: self._on_changed(cb, None))
        else:
            w = QtWidgets.QLabel(f"Unknown control {spec.name}")

        return w

    def _on_changed(self, callback: Callable[[Any], None], value: Any):
        callback(value)

    def _refresh_ui(self):
        # re-fetch specs and update every widget
        def walk(spec: Union[LayoutSpec, ControlSpec]):
            if isinstance(spec, ControlSpec):
                w = self._widgets.get(spec.name)
                if not w:
                    return
                if spec.type == "combo" or spec.type == 'combo-edit':
                    w.blockSignals(True)
                    w.setCurrentText(spec.value)
                    w.blockSignals(False)
                elif spec.type == "checkbox":
                    w.blockSignals(True)
                    w.setChecked(spec.value)
                    w.blockSignals(False)
                elif spec.type == "range_slider":
                    w.blockSignals(True)
                    w.setRange(*(spec.range or (0, 1)))
                    w.setValue(tuple(spec.value))
                    w.blockSignals(False)
                elif spec.type == "slider":
                    w.blockSignals(True)
                    w.setRange(*(spec.range or (0, 1)))
                    w.setValue(spec.value)
                    w.blockSignals(False)
            else:
                for c in spec.children:
                    walk(c)

        root = self.controller.get_layout()
        walk(root)

class ColorMapControls(ColorMapControlsBase):
    controller_class = ColorMapController

class RGBColorControls(ColorMapControlsBase):
    controller_class = RGBMapController

